// Radiohead.cpp : Defines the entry point for the console application.
//

#define TRUE true
#define FALSE false

#include "Utility.h"
#include "Platform.h"
#include "PointCloud.h"
#include "Maths.h"
#include "Perlin.h"

#include "stdio.h"
//#include "glut.h"

#include "fmod.h"

#include <math.h>
#include <iostream>
#include <vector>
#include <algorithm>
//#include <windows.h>

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

using namespace std;

const float kWidth = 640;
const float kHeight = 480;

// camera coordinates
float g_cyaw = 180.0f;
float g_cpitch = 180.0f;
float g_czoom = -650.0f;

float g_focusx = 0.0f;
float g_focusy = 0.0f;
float g_focusz = 0.0f;

// perlin noise scale
float g_noiseAmplitude = 50.0f;
float g_noiseFrequency = 0.0f;

// absolute time of last frame
double g_lastFrameStartTime;
double g_totalElapsedSecs;
int g_currentFrame;

// whether we should sort points or not
bool g_sortPoints = true;
double g_sortTime = 0.0f;

float g_pointSize = 4.0f;
bool g_paused = false;
bool g_mute = false;

int g_lastMousex;
int g_lastMousey;

float g_colour[3] = {0.0f, 0.0f, 1.0f};

// used for debugging to make things faster (there is no other hard limit except for available memory)
const int kMaxFrameCount = 10000;

// all the anim frames
vector<PointCloud*> g_anim;

// FMod mp3 track
FSOUND_SAMPLE* g_sample;

typedef void (*VizFunc)(const PointCloud&);
VizFunc g_visualizer = NULL;

// helper function for converting the csv data to binary (i got sick of loading text all day)
void ConvertToBinary(const char* filename)
{
	// check to see if file already exists, if so don't convert
	FILE* exists = fopen(filename, "rb");
	if (exists)
	{
		cout << "anim.bin already exists, skipping conversion" << endl;
		fclose(exists);
		return;
	}

	cout << "Converting point data to " << filename << endl;

	// open binary file for writing
	FILE* output = fopen(filename, "wb");

	// we know a little about the data, roughly 12000 points per frame, frames start from 1.csv onwards
	vector<Point> points;
	points.reserve(15000);

	int frameCount=1;

	// read files until none left
	for(;;)
	{
		if (frameCount > kMaxFrameCount)
			break;

		// clear out points
		points.resize(0);

		// construct file name
		char fname[255];
		sprintf(fname, "data/%d.csv", frameCount);

		// time from now on
		double startTime = GetSeconds();

		FILE* file = fopen(fname, "r");

		if (file == NULL)
		{
			cout << "Could not open " << fname << " for reading." << endl;
			break;
		}
		else
		{
			// the unfortunate fact is that stdio is just so much easier than iostream for this kind of parsing
			Point p;

			while (!feof(file))
			{
				int intensity;
				fscanf(file, "%f,%f,%f,%d", &p.x, &p.y, &p.z, &intensity);
				
				// splat intensity to rgb
				p.intensity = (intensity<<16)|(intensity<<8)|intensity;

				if (intensity >=25)
					points.push_back(p);
			}

			fclose(file);

			// time to finish
			double endTime = GetSeconds();

			int numPoints = int(points.size());

			// if this file is blank then quit
			if (numPoints == 0)
				break;

			cout << "Convert " << numPoints << " points from " << fname << " in " << endTime-startTime << " seconds" << endl;

			// increase frame count
			++frameCount;

			// write to binary file with a header of how many points in this frame
			fwrite(&numPoints, sizeof(numPoints), 1, output);
			// now the point data
			fwrite(&points[0], sizeof(Point), numPoints, output);
		}
	}

	cout << "Read " << frameCount << " frames" << endl;

	// close binary file
	fclose(output);
}

void LoadFromBinary(const char* filename)
{
	FILE* input = fopen(filename, "rb");

	if (input == NULL)
	{
		cout << "Could not open " << filename << " for reading." << endl;
		return;
	}
	else
	{
		// keep reading frames until we're done
		while (!feof(input) && g_anim.size() < kMaxFrameCount)
		{
			int pointCount=0;

			// time from now on
			double startTime = GetSeconds();

			// read point count
			fread(&pointCount, sizeof(pointCount), 1, input);

			// allocate memory for frame
			Point* points = new Point[pointCount];
			fread(points, sizeof(Point), pointCount, input);	

			// time to finish
			double endTime = GetSeconds();

			// construct new point cloud and add to anim
			PointCloud* pointCloud = new PointCloud(points, pointCount);
			g_anim.push_back(pointCloud);

			// debug info
			cout << "Read frame number " << g_anim.size() << " with " << pointCount << " points in " << endTime-startTime << " seconds" << endl;
		}
	}

	fclose(input);
}


void DrawText(int x, int y, char *string)
{
	int len, i;

	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	int w = glutGet(GLUT_WINDOW_WIDTH);
	int h = glutGet(GLUT_WINDOW_HEIGHT);
	gluOrtho2D(0, w, h, 0);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glColor3f(0.9f, 0.6f, 0.6f);
	glRasterPos2i(x, y);
	len = (int) strlen(string);
	for (i = 0; i < len; i++)
		glutBitmapCharacter(GLUT_BITMAP_9_BY_15, string[i]);

	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
}

void DebugDraw()
{
	char buf[1024];
	int y=20;

	sprintf(buf, "Time %f", g_totalElapsedSecs);
	DrawText(10, y, buf);
	y+=15;

	sprintf(buf, "Frame %d", g_currentFrame);
	DrawText(10, y, buf);
	y+=15;

	sprintf(buf, "Sort time %f", g_sortTime);
	DrawText(10, y, buf);
	y+=15;

	sprintf(buf, "Noise Amplitude: %f", g_noiseAmplitude);
	DrawText(10, y, buf);
	y+=15;

	sprintf(buf, "Noise Frequency: %f", g_noiseFrequency);
	DrawText(10, y, buf);
	y+=15;

}

void ResetTime()
{
	if (g_sample)
	{
		FSOUND_PlaySoundEx(0, g_sample, NULL, TRUE);
		//FSOUND_SetCurrentPosition(0, 0);
	}

	g_currentFrame = 0;
	g_totalElapsedSecs = 0.0;
}

void SimulationLoop()
{
	// if not paused let the sound out
	if (g_sample)
		FSOUND_SetPaused(0, g_paused?TRUE:FALSE);

	glClearColor(0,0,0,0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// some debug text
	DebugDraw();

	PointCloud* p = g_anim[g_currentFrame];

	// update current frame
	g_currentFrame = int(g_totalElapsedSecs * 30.0);
	
	// if looping reset sample and time
	if (g_currentFrame >= g_anim.size())
		ResetTime();

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0.0f, 0.0f, g_czoom);
	glRotatef(g_cyaw, 0, 1, 0);
	glRotatef(g_cpitch, 1, 0, 0);	

	// retrieve the camera position from the view matrix
	float viewMat[16];
	glGetFloatv(GL_MODELVIEW_MATRIX, viewMat);
	float camerax = -g_czoom*viewMat[8] + g_focusx;
	float cameray = -g_czoom*viewMat[9] + g_focusy;
	float cameraz = -g_czoom*viewMat[10] + g_focusz;

	// sort cloud with respect to camera
	if (g_sortPoints)
	{
		double start=GetSeconds();
		p->Sort(camerax, cameray, cameraz);
		double stop=GetSeconds();

		g_sortTime = stop-start;
	}	

	glTranslatef(-g_focusx, -g_focusy, -g_focusz);

	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_SRC_ALPHA);

	// draw point cloud	
	//glEnable(0x8861);
	glEnable(GL_POINT_SMOOTH);	

	/*
	// draw previous points first in blue
	glEnableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
	glPointSize(g_pointSize*0.5f);
	glColor4f(0.0,0.0,1.0,1.0);
	glVertexPointer(3, GL_FLOAT, sizeof(Point), prev->GetPoints());
	glDrawArrays(GL_POINTS, 0, prev->GetNumPoints());
	*/

	/*
	// draw lines connecting the two frames
	glBegin(GL_LINES);
	int minCount = min(prev->GetNumPoints(), p->GetNumPoints());
	glColor4f(0.0f, 1.0f, 0.0f, 1.0f);
	const Point* prevPoints = prev->GetPoints();
	const Point* currPoints = p->GetPoints();

	for (int i=0; i < minCount; ++i)
	{
		glVertex3f(prevPoints->x, prevPoints->y, prevPoints->z);
		glVertex3f(currPoints->x, currPoints->y, currPoints->z);

		++prevPoints;
		++currPoints;
	}
	glEnd();
	*/

	/*	
	// draw current points slightly differently
	glEnableClientState(GL_COLOR_ARRAY);
	glEnableClientState(GL_VERTEX_ARRAY);	
	glPointSize(g_pointSize);
	glVertexPointer(3, GL_FLOAT, sizeof(Point), p->GetPoints());
	glColorPointer(4, GL_UNSIGNED_BYTE, sizeof(Point), &p->GetPoints()->intensity);
	glDrawArrays(GL_POINTS, 0, p->GetNumPoints());
	*/
	
	// call current viz routine
	g_visualizer(*p);

	// flip
	glutSwapBuffers();

	// update timers
	if (!g_paused)
		g_totalElapsedSecs += GetSeconds() - g_lastFrameStartTime;
	
	g_lastFrameStartTime = GetSeconds();
}

void Visualize1(const PointCloud& p)
{
	const Point* currPoints = p.GetPoints();

	glDisableClientState(GL_VERTEX_ARRAY);	
	glDisableClientState(GL_COLOR_ARRAY);	

	glPointSize(g_pointSize);
	glBegin(GL_POINTS);

	for (int i=0; i < p.GetNumPoints(); ++i)
	{
		float xnoise = g_noiseAmplitude*Perlin2D(currPoints->y*g_noiseFrequency, g_totalElapsedSecs, 1, 1);
		float intensity = ((byte*)&currPoints->intensity)[0] * (1.0f/255.0f);

		glColor4f(intensity, intensity, intensity, 0.0f);
		glVertex3f(currPoints->x+xnoise, currPoints->y, currPoints->z);

		++currPoints;
	}

	glEnd();
}

void Visualize2(const PointCloud& p)
{
	const Point* currPoints = p.GetPoints();

	glDisableClientState(GL_VERTEX_ARRAY);	
	glDisableClientState(GL_COLOR_ARRAY);	

	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_SRC_ALPHA);
	glPointSize(g_pointSize);
	glBegin(GL_LINES);

	static float s_prevNoiseAmplitude=g_noiseAmplitude;
	static float s_prevNoiseFrequency=g_noiseFrequency;
	static float s_prevTotalElapsedSecs=g_totalElapsedSecs;

	const float kRed[3] = { 1.0f, 0.0f, 0.0f };
	const float kGreen[3] = {0.0f, 1.0f, 0.0f };
	const float kBlue[3] = { 0.0f, 0.0f, 1.0f };

	const float kRainbow[5][3] = { {1.0f, 0.0f, 0.0f},
								   {1.0f, 1.0f, 0.0f},
								   {0.0f, 1.0f, 0.0f},
								   {0.0f, 1.0f, 1.0f},
								   {0.0f, 0.0f, 1.0f},
								 };
	
	for (int i=0; i < p.GetNumPoints(); ++i)
	{
		float prevxnoise = 0.8*s_prevNoiseAmplitude*Perlin2D(currPoints->y*s_prevNoiseFrequency, s_prevTotalElapsedSecs, 1, 1);
		float xnoise = g_noiseAmplitude*Perlin2D(currPoints->y*g_noiseFrequency, g_totalElapsedSecs, 1, 1);
	
		float intensity = ((byte*)&currPoints->intensity)[0] * (1.0f/255.0f);

		glColor4f(0, 0, intensity, 0.5f);
		glVertex3f(currPoints->x+xnoise, currPoints->y, currPoints->z);

		float p = fmod(abs(currPoints->y)*g_noiseFrequency, 5.0f);
		float t = p;
		int c1 = int(t);
		int c2 = (c1+1)%5;
		float f = t - c1;

		glColor4f(Lerp(kRainbow[c1][0], kRainbow[c2][0], f)*intensity*2,
				  Lerp(kRainbow[c1][1], kRainbow[c2][1], f)*intensity*2,
				  Lerp(kRainbow[c1][2], kRainbow[c2][2], f)*intensity*2, 1.0f);

		glVertex3f(currPoints->x+prevxnoise, currPoints->y, currPoints->z+10.0f);

		++currPoints;
	}

	glEnd();

	// update previous states
	s_prevNoiseAmplitude=g_noiseAmplitude;
	s_prevNoiseFrequency=g_noiseFrequency;
	s_prevTotalElapsedSecs=g_totalElapsedSecs;
}

void Keyboard(unsigned char key, int x, int y)
{
	switch (key)
	{
	case 27:
	exit(0);
	break;
	case '1':
		g_visualizer = Visualize1;
		break;
	case '2':
		g_visualizer = Visualize2;
		break;
	case '3':
		break;
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	case 's':
	{
		g_mute = !g_mute;
		FSOUND_SetMute(0, g_mute?TRUE:FALSE);
		break;
	}
	case 'r':
		ResetTime();
		break;
	case ' ':
		g_paused = !g_paused;
		break;
	}
}

void ArrowKeys(int key, int x, int y)
{
	switch (key)
	{
	case GLUT_KEY_LEFT:
		break;
	case GLUT_KEY_RIGHT:

		break;
	case GLUT_KEY_UP:
		g_pointSize += 0.1f;
		break;
	case GLUT_KEY_DOWN:
		g_pointSize -= 0.1f;
		break;
	}
}

void MouseUpdatePosition(int x, int y)
{
	g_lastMousex = x;
	g_lastMousey = y;
}

void MouseZoomFunc(int x, int y)
{
	g_czoom += y-g_lastMousey;

	MouseUpdatePosition(x, y);
}

void MouseRotateFunc(int x, int y)
{
	g_cyaw += x - g_lastMousex;
	g_cpitch -= y - g_lastMousey;

	MouseUpdatePosition(x, y);
}

void MouseNoiseFunc(int x, int y)
{
	g_noiseAmplitude = max(0.0f, g_noiseAmplitude + x - g_lastMousex);
	g_noiseFrequency = max(0.0f, g_noiseFrequency - (y - g_lastMousey)*0.0001f);
	
	MouseUpdatePosition(x, y);
}

void MouseFunc(int button, int state, int x, int y)
{
	if (state == GLUT_DOWN)
	{
		if (button == GLUT_LEFT_BUTTON)
			glutMotionFunc(MouseRotateFunc);
		
		if (button == GLUT_MIDDLE_BUTTON)
			glutMotionFunc(MouseNoiseFunc);

		if (button == GLUT_RIGHT_BUTTON)
			glutMotionFunc(MouseZoomFunc);
	}
}

void Reshape(int width, int height)
{
	glViewport(0, 0, width, height);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45, float(width)/height, 0.1, 10000);
}

void Init()
{
	// only converts if file does not already exist
	ConvertToBinary("data/anim.bin");

	// load anim data
	LoadFromBinary("data/anim.bin");

	// calculate mean position, note this is purposely not the 
	// center of the bounds as there are a lot of outliers
	if (g_anim.size())
	{
		PointCloud* cloud = g_anim[0];
		const Point* points = cloud->GetPoints();

		float rcpCount = 1.0f / cloud->GetNumPoints();

		for (int i=0; i < cloud->GetNumPoints(); ++i)
		{
			const Point& p = points[i];

			g_focusx += p.x*rcpCount;
			g_focusy += p.y*rcpCount;
			g_focusz += p.z*rcpCount;
		}

		cout << "Center at " << g_focusx << "," << g_focusy << "," << g_focusz << endl;
	}

	// initialize timer
	g_lastFrameStartTime = GetSeconds();

	// set default visualizer
	g_visualizer = Visualize2;
}

void DeInit()
{
	for_each(g_anim.begin(), g_anim.end(), free_ptr<PointCloud>());
}

int main(int argc, char** argv)
{
	// init fmod first
	if (FSOUND_Init(44100, 32, 0) == TRUE)
	{
		g_sample = FSOUND_Sample_Load(0, "data/HouseOfCards_DataSample.mp3", FSOUND_NORMAL, 0, 0);

		if (g_sample)
			FSOUND_PlaySoundEx(0, g_sample, NULL, TRUE);
		else
			cout << "Couldn't load sample into fsound" << endl;
	}
	else
		cout << "Could not initialise FMod" << endl;

	// init gl
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH );

	glutInitWindowSize(kWidth, kHeight);
	glutCreateWindow("House Of Cards");

	Init();

	glutMouseFunc(MouseFunc);
	glutPassiveMotionFunc(MouseUpdatePosition);
	glutReshapeFunc(Reshape);
	glutDisplayFunc(SimulationLoop);
	glutKeyboardFunc(Keyboard);
	glutSpecialFunc(ArrowKeys);
	glutIdleFunc(SimulationLoop);

	glutMainLoop();

	// free up memory
	DeInit();

	// close fmod
	FSOUND_Close();
	
	return 0;
}
