#include "Utility.h"

#include "Maths.h"

#include <iostream>
#include <cassert>


//---------------------------------------------------------------
// class used to evaluate functions over a specified range

RangeValue::RangeValue(float start, float stop, Function f) : m_start(start)
															, m_stop(stop)
															, m_func(f)
{
}

void RangeValue::Set(float start, float stop, Function f)
{
	m_start = start;
	m_stop = stop;
	m_func = f;
}

// evaluate the range given an interpolant 0-1 to be valid
float RangeValue::Evaluate(float t) const
{
	assert(t >= 0.0f && t <= 1.0f);

	// linearly interpolate 
	float v = Lerp(m_start, m_stop, t);

	// take function of
	switch (m_func)
	{
	case kFuncNone:
		return v;
	case kFuncCos:
		return cosf(v);
	case kFuncSin:
		return sinf(v);
	case kFuncTan:
		return tanf(v);
	case kFuncArcCos:
		return acosf(v);
	case kFuncArcSin:
		return asinf(v);
	case kFuncArcTan:
		return atanf(v);
	case kFuncSinh:
		return sinh(v);
	case kFuncCosh:
		return cosh(v);
	default:
		return v;
	};
}

std::istream& operator >> (std::istream& s, RangeValue& v)
{
	return s >> v.m_start >> v.m_stop;

	// TODO: deserialize function type
}

std::ostream& operator << (std::ostream& s, const RangeValue& v)
{
	return s << v.m_start << v.m_stop;

	// TODO: serialize function type
}
