#pragma once

#include <iosfwd>
#include <functional>

class RangeValue 
{
public:

	enum Function
	{
		kFuncNone,
		kFuncCos,
		kFuncSin,
		kFuncTan,
		kFuncArcSin,
		kFuncArcCos,
		kFuncArcTan,
		kFuncSinh,
		kFuncCosh		
	};

	// constructor, by default just does a linear interpolation and returns
	RangeValue(float start, float stop, Function f=kFuncNone);
	
	// reset values
	void Set(float start, float stop, Function f=kFuncNone);

	// evaluate the range given an interpolant 0-1 to be valid
	float Evaluate(float t) const;

private:

	float m_start;
	float m_stop;

	Function m_func;

	friend std::istream& operator >> (std::istream& s, RangeValue& v);
	friend std::ostream& operator << (std::ostream& s, const RangeValue& v);
};

std::istream& operator >> (std::istream& s, RangeValue& v);
std::ostream& operator << (std::ostream& s, const RangeValue& v);

// growth functions

// returns the growth rate given by Velhurst's logistic function 
inline float SigmoidGrowth(float x, float xmax)
{
	if (xmax == 0.0f)
		return 0.0f;

	return x*(1.0f-x/xmax);
}

// functor designed for use in the stl
template <typename T>
class free_ptr : public std::unary_function<T*, void>
{
public:

	void operator()(const T* ptr)
	{
		delete ptr;
	}
};